@@ -813,7 +813,8 @@ unsigned m68k_read_memory_8(unsigned address) {
                    Bit 6:    0 = 2 slots, 1 = 4 or 6 slots
                    Bits 0-5: Unknown
                 */
-                return geo_input_sys_cb[2](); // Active Low
+                // Only 1 and 2 slot System ROMs are supported, therefore ~0x40
+                return geo_input_sys_cb[2]() & ~0x40; // Active Low
             }
             case 0x320000: { // REG_SOUND
                 return ngsys.sound_reply; // Z80 Reply Code
@@ -828,7 +829,8 @@ unsigned m68k_read_memory_8(unsigned address) {
                    Bit 1:    Coin-in 2
                    Bit 0:    Coin-in 1
                 */
-                return geo_input_sys_cb[0]() | (geo_rtc_rd() << 6);
+                // Coin slots 3 and 4 are never used, therefore | 0x18
+                return geo_input_sys_cb[0]() | (geo_rtc_rd() << 6) | 0x18;
             }
             case 0x340000: { // REG_P2CNT
                 return geo_input_cb[1](1);
